unit TreeStrings;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Contnrs, LazUtf8Classes;

type

  { TTreeStringsItem }

  TTreeStringsItem = class
  private
    FParent: TTreeStringsItem;
    FCaption: string;
    FStrings: TStrings;
    FItems: TObjectList;
    function GetCount: Integer;
    function GetItems(Index: Integer): TTreeStringsItem;
    procedure SetParent(AValue: TTreeStringsItem);
  protected
  public
    constructor Create; virtual;
    destructor Destroy; override;
    function IsThere(AItem: TTreeStringsItem): Boolean;
    procedure Clear;
    function AddChild(AChildCaption: string): TTreeStringsItem;
    function CopyAsChild(AItem: TTreeStringsItem): TTreeStringsItem;
    function CopyAsSibling(AItem: TTreeStringsItem): TTreeStringsItem;
    function MoveAsChild(AItem: TTreeStringsItem): TTreeStringsItem;
    function MoveAsSibling(AItem: TTreeStringsItem): TTreeStringsItem;
    function AddNext(AChildCaption: string): TTreeStringsItem;
    function Delete(AChild: TTreeStringsItem): Boolean;
    function MoveUp(): Boolean;
    function MoveDown(): Boolean;
    function MoveTop(): Boolean;
    function MoveBottom(): Boolean;
  public
    property Parent: TTreeStringsItem read FParent write SetParent;
    property Caption: string read FCaption write FCaption;
    property Count: Integer read GetCount;
    property Strings: TStrings read FStrings;
    property Items[Index: Integer]: TTreeStringsItem read GetItems;
  end;

  TTreeStrings = class(TTreeStringsItem)
  private
    FBeginStr: string;
    FEndStr: string;
  protected
  public
    constructor Create; override;
    destructor Destroy; override;
    procedure LoadFromStrings(AStrings: TStrings);
    procedure SaveToStrings(AStrings: TStrings; AEndStr: Boolean {= FALSE});
  end;

implementation

uses
  Unit1;

{ TTreeStrings }

constructor TTreeStrings.Create;
begin
  inherited Create;
  FBeginStr := '<<';
  FEndStr := '>>';
end;

destructor TTreeStrings.Destroy;
begin
  inherited Destroy;
end;

procedure TTreeStrings.LoadFromStrings(AStrings: TStrings);
var
  Item: TTreeStringsItem;
  I: Integer;
  S: string;
  function _Check(SMark: string; S: string): Boolean; inline;
  begin
    Result := SMark + ' ' = Copy(S, 1, Length(SMark) + 1);
  end;
begin
  Clear;
  Item := Self;
  for I := 0 to AStrings.Count - 1 do
  begin
    S := AStrings[I];

    if (FBeginStr + ' ' = Copy(S, 1, Length(FBeginStr) + 1)) and
       (Length(S) > Length(FBeginStr) + 1) then
      Item := Item.AddChild(Trim(Copy(S, Length(FBeginStr) + 1, MaxInt)))
    else if FEndStr = Copy(S, 1, Length(FEndStr)) then
    begin
      if Assigned(Item.Parent) then
        Item := Item.Parent;
    end
    else
    begin
      // 行頭の２重化された項目開始マーク or 終了マークは、
      // エスケープシーケンス処理された文字列とみなして復元する（１重化）
      if FBeginStr + FBeginStr = Copy(S, 1, Length(FBeginStr) * 2) then
        System.Delete(S, 1, Length(FBeginStr))
      else if FEndStr + FEndStr = Copy(S, 1, Length(FEndStr) * 2) then
        System.Delete(S, 1, Length(FEndStr));
      Item.FStrings.Add(S);
    end;
  end;
end;

procedure TTreeStrings.SaveToStrings(AStrings: TStrings; AEndStr: Boolean);
var
  I: Integer;
  procedure _Add(S: string); inline;
  begin
    // 行頭の項目開始マーク or 終了マークを２重化する
    // （エスケープシーケンス処理）
    if FBeginStr = Copy(S, 1, Length(FBeginStr)) then
      S := FBeginStr + S
    else if FEndStr = Copy(S, 1, Length(FEndStr)) then
      S := FEndStr + S;
    AStrings.Add(S);
  end;
  procedure _Save(AItem: TTreeStringsItem);
  var
    I: Integer;
  begin
    AStrings.Add(FBeginStr + ' ' + AItem.Caption);

    for I := 0 to AItem.FStrings.Count - 1 do
      _Add(AItem.FStrings[I]);
    for I := 0 to AItem.FItems.Count - 1 do
      _Save(TTreeStringsItem(AItem.FItems[I]));

    if AEndStr then
      AStrings.Add(FEndStr + ' ' + AItem.Caption)
    else
      AStrings.Add(FEndStr);
  end;
begin
  for I := 0 to Self.FStrings.Count - 1 do
    _Add(Self.FStrings[I]);
  for I := 0 to Self.FItems.Count - 1 do
    _Save(TTreeStringsItem(Self.FItems[I]));
end;

{ TTreeStringsItem }

function TTreeStringsItem.GetItems(Index: Integer): TTreeStringsItem;
begin
  Result := TTreeStringsItem(FItems[Index]);
end;

procedure TTreeStringsItem.SetParent(AValue: TTreeStringsItem);
var
  Item: TTreeStringsItem;
  Is_NewParent_Descendant: Boolean;
begin
  if FParent = AValue then Exit;

  // 子孫を Parent に設定しないように
  Is_NewParent_Descendant := FALSE;
  Item := AValue.Parent;
  while Assigned(Item) do
  begin
    if Item = Self then
    begin
      Is_NewParent_Descendant := TRUE;
      break;
    end;
    Item := Item.Parent;
  end;
  if Is_NewParent_Descendant then
    //raise ''
    Exit;

  if Assigned(FParent) then FParent.FItems.Extract(Self);
  FParent := AValue;
  if Assigned(FParent) then FParent.FItems.Add(Self);
end;

function TTreeStringsItem.GetCount: Integer;
begin
  Result := FItems.Count;
end;

constructor TTreeStringsItem.Create;
begin
  FStrings := LazUtf8Classes.TStringListUTF8.Create;
  FItems := TObjectList.Create(TRUE);
end;

destructor TTreeStringsItem.Destroy;
begin
  FItems.Free;
  FStrings.Free;
  inherited Destroy;
end;

procedure TTreeStringsItem.Clear;
begin
  FCaption := '';
  FStrings.Clear;
  FItems.Clear;
end;

function TTreeStringsItem.AddChild(AChildCaption: string): TTreeStringsItem;
begin
  Result := TTreeStringsItem.Create;

  //Result.SetParent(Self);
  Result.FParent := Self;
  FItems.Add(Result);

  Result.Caption := AChildCaption;
end;

function TTreeStringsItem.CopyAsChild(AItem: TTreeStringsItem): TTreeStringsItem;
var
  I: Integer;
begin
  Result := TTreeStringsItem.Create;
  try
    Result.FCaption := AItem.Caption;
    Result.FStrings.Assign(AItem.FStrings);

    for I := 0 to AItem.FItems.Count - 1 do
      Result.CopyAsChild(AItem.Items[I]);

  finally
    Result.FParent := Self;
    FItems.Add(Result);
  end;
end;

function TTreeStringsItem.CopyAsSibling(AItem: TTreeStringsItem): TTreeStringsItem;
var
  I: Integer;
begin
  Result := TTreeStringsItem.Create;
  try
    Result.FCaption := AItem.Caption;
    Result.FStrings.Assign(AItem.FStrings);

    for I := 0 to AItem.FItems.Count - 1 do
      Result.CopyAsChild(AItem.Items[I]);

  finally
    Result.SetParent(Self.Parent);
  end;
end;

function TTreeStringsItem.MoveAsChild(AItem: TTreeStringsItem): TTreeStringsItem;
begin
  AItem.SetParent(Self);
  if AItem.Parent = Self then
    Result := AItem
  else
    Result := nil;
end;

function TTreeStringsItem.MoveAsSibling(AItem: TTreeStringsItem): TTreeStringsItem;
begin
  AItem.SetParent(Self.Parent);
  if AItem.Parent = Self.Parent then
    Result := AItem
  else
    Result := nil;
end;

function TTreeStringsItem.AddNext(AChildCaption: string): TTreeStringsItem;
var
  I: Integer;
begin
  if Assigned(FParent) then
  begin
    I := FParent.FItems.IndexOf(Self);
    if I >= 0 then
    begin
      Result := TTreeStringsItem.Create;
      Result.FParent := FParent;
      FParent.FItems.Insert(I + 1, Result);
      Result.Caption := AChildCaption;
    end;
  end;
end;

function TTreeStringsItem.Delete(AChild: TTreeStringsItem): Boolean;
var
  I: Integer;
begin
  I := FItems.IndexOf(AChild);
  if I >= 0 then
  begin
    FItems.Delete(I);
    Result := TRUE;
  end
  else
    Result := FALSE;
end;

function TTreeStringsItem.MoveUp(): Boolean;
var
  I: Integer;
begin
  Result := FALSE;
  if Assigned(FParent) then
  begin
    I := FParent.FItems.IndexOf(Self);
    if I > 0 then
    begin
      FParent.FItems.Move(I, I - 1);
      Result := TRUE;
    end;
  end;
end;

function TTreeStringsItem.MoveDown(): Boolean;
var
  I: Integer;
begin
  Result := FALSE;
  if Assigned(FParent) then
  begin
    I := FParent.FItems.IndexOf(Self);
    if(I >= 0)and(I < (FParent.FItems.Count - 1)) then
    begin
      FParent.FItems.Move(I, I + 1);
      Result := TRUE;
    end;
  end;
end;

function TTreeStringsItem.MoveTop(): Boolean;
var
  I: Integer;
begin
  Result := FALSE;
  if Assigned(FParent) then
  begin
    I := FParent.FItems.IndexOf(Self);
    if I > 0 then
    begin
      FParent.FItems.Move(I, 0);
      Result := TRUE;
    end;
  end;
end;

function TTreeStringsItem.MoveBottom(): Boolean;
var
  I: Integer;
begin
  Result := FALSE;
  if Assigned(FParent) then
  begin
    I := FParent.FItems.IndexOf(Self);
    if(I >= 0)and(I < (FParent.FItems.Count - 1)) then
    begin
      FParent.FItems.Move(I, FParent.FItems.Count - 1);
      Result := TRUE;
    end;
  end;
end;

function TTreeStringsItem.IsThere(AItem: TTreeStringsItem): Boolean;
var
  I: Integer;
begin
  if AItem = nil then
    Result := FALSE
  else if AItem = Self then
    Result := TRUE
  else
  begin
    for I := 0 to FItems.Count - 1 do
    begin
      Result := TTreeStringsItem(FItems[I]).IsThere(AItem);
      if Result then
        break;
    end;
  end;
end;

end.

